#!/usr/bin/python3

""" Cloud Backup Agent Updater """

import logging
import socket
import sys
import hashlib
import inspect
import io
import shutil
import subprocess
import tarfile
import re
import os
import pdb
from argparse import ArgumentParser
from collections import namedtuple, ChainMap
from functools import partial, lru_cache, total_ordering, wraps
from itertools import chain, groupby
from logging import INFO, DEBUG, WARNING, ERROR, NOTSET
from platform import python_version
from textwrap import dedent, fill
from configparser import ConfigParser
from operator import attrgetter
from os.path import (expanduser, basename, splitext, abspath,
                     getsize, relpath, normpath, realpath, dirname)
from subprocess import CalledProcessError
from tarfile import CompressionError
from tempfile import TemporaryDirectory
from types import SimpleNamespace
from urllib.request import urlopen, HTTPError, URLError
from urllib.parse import urlparse

try:
    from enum import Enum
except ImportError:
    def Enum(clsname, names=None):  # pylint: disable=invalid-name
        """ Enum replacement for py<3.4

        Acts as stock Enum.__call__, except not all arguments are supported.
        Values not given are assigned to a new object().
        """
        if isinstance(names, str):
            # string to dict
            names = {name: object() for name in names.split()}
        names = dict(names)  # kv-pairs to dict, or no-op if already dict
        return type(clsname, (), names)


class CbauLogger(logging.getLoggerClass()):
    """ Logger subclass supporting a PRINT pseudo-loglevel and method

    Some commands have no pipe-able payload, but print status messages as part
    of normal operation. Some such messages should be logged, but should also
    be printed 'almost' all the time, which most log messages are not. Existing
    loglevels are inappropriate; INFO isn't printed by default, and
    WARNING/ERROR would imply problems where there are none.

    The 'PRINT' loglevel logs the message at a relatively high level (halfway
    between warning and error), ensuring that it actually shows up on the
    console, but without the implication that something is wrong.
    """

    # Create a PRINT loglevel between ERROR and WARNING
    # These messages should always be printed unless --quiet is given.
    PRINT = (logging.ERROR + logging.WARNING) // 2

    fmt = SimpleNamespace()
    fmt.simple = '%(message)s'
    fmt.leveled = '%(levelname)s\t%(message)s'
    fmt.detailed = '%(levelname)s\t%(module)s:%(lineno)s\t%(message)s'
    fmt.date = '%Y-%m-%dT%H:%M:%S'

    def print(self, msg, *args, **kwargs):
        """ Log a message at the PRINT pseudo-loglevel

        This method behaves just like the other log.{level} methods. Use it to
        unconditionally print something and also ensure that it is logged.
        """
        self.log(self.PRINT, msg, *args, **kwargs)

    @classmethod
    def setup(cls):
        """ Configure the logging system to use CbauLogger and return it """
        logging.addLevelName(cls.PRINT, 'PRINT')
        logging.setLoggerClass(cls)
        cls.configure()
        return logging.getLogger(__name__)

    @classmethod
    def configure(cls, args=None, conf=None):
        """ Configure the logging handlers for the root logger """
        # We want to enforce the loglevel by handler rather than by logger, so
        # that console and file loglevels can be controlled separately. The
        # root log's level is WARNING by default, which will mess with that.
        # Fix it up before returning our logger.
        root = logging.getLogger()
        root.setLevel(NOTSET)
        for handler in root.handlers:
            root.removeHandler(handler)
        cls._add_console_handler(root, args)
        if conf:
            cls._add_file_handler(root, conf)

    @classmethod
    def _add_console_handler(cls, logger=None, args=None):
        """ Set up console output """
        logger = logger or logging.getLogger()
        args = args or SimpleNamespace(debug=False, verbose=False, quiet=False)
        level = (DEBUG if args.debug
                 else INFO if args.verbose or getattr(args, 'dryrun', None)
                 else ERROR if args.quiet
                 else WARNING)
        fmt = (cls.fmt.detailed if args.debug
               else cls.fmt.leveled if args.verbose
               else cls.fmt.simple)
        console = logging.StreamHandler()
        console.setLevel(level)
        console.setFormatter(logging.Formatter(fmt))
        logger.addHandler(console)
        logger.debug("console initialized with fmt: %s", fmt)

    @classmethod
    def _add_file_handler(cls, logger=None, conf=None):
        logger = logger or logging.getLogger()
        path = expanduser(conf.log_file)
        level = conf.log_level.upper()
        formatter = logging.Formatter(conf.log_format, conf.log_datefmt)
        try:
            logfile = logging.FileHandler(path)
        except OSError as ex:
            log.warning("failed to open '%s' for logging; "
                        "logging disabled (%s)", path, ex)
            return
        logfile.setFormatter(formatter)
        logfile.setLevel(level)
        logger.addHandler(logfile)


log = CbauLogger.setup()
NO_DEFAULT = object()  # alternate default arg when None is inappropriate
Mode = Enum('Mode', 'FORCE DRYRUN INTERACTIVE')


# NOTE: For now, we're trying to keep the user-side updater script as a single
# file with no external dependencies. I've tried to organize it in sections in
# the hopes of preventing it from turning into complete chaos.


# EXCEPTIONS


class CbauError(Exception):
    """ Base class for updater exceptions """


class ValidationFailed(CbauError):
    """ General validation failures """


class ValidationError(ValidationFailed):
    """ Exception indicating a downloaded package failed validation """
    def __init__(self, check, expected, actual):
        self.check = check
        self.expected = expected
        self.actual = actual
        msg = "{} validation failed: expected {}, got {}"
        super().__init__(msg.format(check, expected, actual))


class InvalidPackageFilename(CbauError, ValueError):
    """ Exception indicating a file isn't a package """
    def __init__(self, filename):
        self.filename = filename
        msg = "Not a package file: {}".format(filename)
        super().__init__(msg)


class BrokenRepo(CbauError):
    """ Exception indicating a problem with the package repository """


class UserAbort(CbauError):
    """ Exception for when a confirmation prompt is rejected """


class OperationFailed(CbauError):
    """ General failure """


# UTILITY FUNCTIONS

def getversion():
    """ Get cbau's version

    For now, the returned value is a string, not a Version object, because
    Version's ontology doesn't quite match that used by pip.

    The version is acquired differently depending on how cbau is installed.
    setuptools-scm puts it in version.py. The tar package puts it in a text
    file in the share directory.

    If there is no installation, we can't figure out the version.
    """
    try:
        log.debug("looking for version in version.py")
        from .version import version  # pylint: disable=import-outside-toplevel
        return version
    except (ImportError, SystemError) as ex:
        # Note, systemerror needs catching because of
        # https://bugs.python.org/issue18018
        pass
    path = '{}/../share/version.txt'.format(dirname(realpath(__file__)))
    try:
        log.debug("looking for version in %s", path)
        with open(path, encoding='utf-8') as f:
            return f.read().strip()
    except OSError:
        pass
    return 'unknown (updater.py hash {})'.format(filehash(__file__))


def eprint(*args, **kwargs):
    """ Print to stderr """
    print(*args, file=sys.stderr, **kwargs)


def logquit(errcode, *args, **kwargs):
    """ Convenience function for log-and-quit """
    log.log(*args, **kwargs)
    sys.exit(errcode)


# Convenience shorthands for e.g. 'do_thing() or abort()'
halt = partial(logquit, 0, logging.INFO)
abort = partial(logquit, 1, logging.ERROR)
fail = partial(logquit, 1, logging.CRITICAL)


def find(filename, path, default=NO_DEFAULT):
    """ Find a file under a path

    Convenience shorthand for locating files after package extraction, since
    there's no direct mapping from tarball paths to install paths. Exact
    matching only. Returns the path to the first file found, or None if there
    is no match.
    """
    log.debug("looking for %s under %s", filename, path)
    for root, _, files in os.walk(path):
        for entry in files:
            if entry == filename:
                return os.path.join(root, entry)
    if default is NO_DEFAULT:
        raise FileNotFoundError(path)
    return default


def filehash(path, hashfunc=hashlib.sha256, chunk_blocks=256):
    """ Get the hash of a file as a hex string """
    hasher = hashfunc()
    sz_chunks = chunk_blocks * hasher.block_size
    with open(path, 'rb') as f:
        chunk = f.read(sz_chunks)
        while chunk:
            hasher.update(chunk)
            chunk = f.read(sz_chunks)
    return hasher.hexdigest()


@lru_cache(128)
def available_compression_formats():
    """ Figure out what compression formats are available

    Old systems may be missing bz2/lzma. We want this information so we don't
    download something we can't subsequently unpack.
    """
    available = []
    for ext in 'gz', 'bz2', 'xz':
        fmode = 'w:' + ext
        try:
            with tarfile.open(os.devnull, fmode):
                log.debug("%s compression is available", ext)
                available.append(ext)
        except CompressionError as ex:
            log.debug(ex)
    log.debug("available compression formats: %s", ', '.join(available))
    return available


# CLASSES


class Action:
    """ Decorator for functions that respond to run mode

    Actions will be skipped in dryrun mode. If `prompt` is given, they will
    also prompt for confirmation in interactive mode.
    """
    runmode = Mode.INTERACTIVE

    def __init__(self, msg, prompt=None, lvl=INFO, rval=None):
        self._rval = rval
        self._msg = msg
        self._loglevel = lvl
        self._prompt = prompt

    @classmethod
    def confirm(cls, prompt, *args):
        """ Present a confirmation prompt

        If args are given, they are percent-formatted into the prompt.
        """
        if cls.runmode is not Mode.INTERACTIVE:
            return True
        prompt = (prompt % args) + " [y/n] "
        eprint(prompt, end='')
        accepted = input().lower().strip() in ['y', 'yes']
        if not accepted:
            raise UserAbort("confirmation declined; aborting")
        return True

    @property
    def msg(self):
        """ Get appropriate log message for current runmode """
        if self.runmode is Mode.DRYRUN:
            return '(not) {} (dry run)'.format(self._msg)
        return self._msg

    @property
    def dryrun(self):
        """ Check if dryrun is active """
        return self.runmode is Mode.DRYRUN

    @property
    def interactive(self):
        """ Check if mode is interactive """
        return self.runmode is Mode.INTERACTIVE

    def _log(self, *args):
        log.log(self._loglevel, self.msg, *args)

    def __call__(self, function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            if self._prompt and self.interactive:
                self.confirm(self._prompt, *args)
            self._log(*args)
            if self.dryrun:
                return self._rval
            return function(*args, **kwargs)
        return wrapper

    @classmethod
    def setrunmode(cls, args):
        """ Set the global run-mode based on parsed arguments """
        def getflag(flag):
            return getattr(args, flag, False)
        cls.runmode = (Mode.DRYRUN if getflag('dryrun')
                       else Mode.INTERACTIVE if getflag('interactive')
                       else Mode.FORCE if getflag('force')
                       else Mode.INTERACTIVE)
        log.debug("run mode set: %s", cls.runmode)


# IMPLEMENTATION WRAPPERS
#
# The below definitions wrap stdlib functions such that the result obeys
# dryrun and handles some edge cases. The intent is to make dryrun show what's
# actually going to happen during install/uninstall, rather than just noting
# that install/uninstall won't be done.

@Action("installing link: %s -> %s")
def link(path, target):
    """ Force-create a symlink in a sort-of-atomic fashion

    os.symlink has no 'force' option, and remove-replace introduces a period
    where the link isn't there. This function mimics ln -sf by creating a
    temporary link and then renaming it over the intended one.
    """
    tmp = target + '.tmp'
    try:
        os.remove(tmp)
    except FileNotFoundError:
        pass
    os.symlink(target, tmp)
    os.rename(tmp, path)


@Action("moving: %s -> %s")
def move(src, dest):
    """ Move src to dest, replacing any existing dest

    This attempts to be 'safe', in that if moving the directory fails, the
    previously-existing directory is restored.

    Returns the destination path.
    """
    old = dest + '.old'
    try:
        os.rename(dest, old)
    except FileNotFoundError:  # e.g. dest doesn't exist
        pass
    try:
        shutil.move(src, dest)
    finally:
        if not os.path.exists(dest):
            os.rename(old, dest)
        try:
            shutil.rmtree(old)
        except FileNotFoundError:
            pass


@Action("copying: %s -> %s")
def copy(src, dest):
    """ shutil.copy wrapper that obeys dryrun """
    try:
        os.remove(dest)
    except FileNotFoundError:
        pass  # this is fine
    shutil.copy(src, dest)


@Action("removing: %s")
def remove(path):
    """ os.remove wrapper that catches and logs unimportant errors """
    try:
        os.remove(path)
        log.debug("removed: %s", path)
    except FileNotFoundError as ex:
        log.debug("not removing %s (%s)", path, ex)
    except OSError as ex:
        log.warning("couldn't remove %s (%s)", path, ex)


@Action("removing directory: %s")
def rmtree(path):
    """ shutil.rmtree wrapper that obeys dryrun """
    shutil.rmtree(path)


@Action("executing: %s")
def run(*args, **kwargs):
    """ subproccess.check_output wrapper that obeys dryrun

    By default this runs the subproc with shell=True and stderr=sp.STDOUT
    because that's what we most frequently use, but those (and any other args
    for check_output) can be overridden with kwargs.
    """
    sp = subprocess  # convenience alias
    kwargs = ChainMap(kwargs, dict(shell=True, stderr=sp.STDOUT))
    return sp.check_output(*args, **kwargs)


def unpack(tarball, dest):
    """ Unpack a tarball to a given directory

    Note that unpack is deliberately not wrapped as an Action; if it was, there
    would be nothing for further operations to iterate over, hence no way for
    dryrun to show 'what would be done' during an install.

    To ensure unpacking results in a no-op under dryrun conditions, provide a
    TemporaryDirectory as dest, make a PackageContents from it, and then move
    it to its intended destination. move() *is* an Action, so it will get
    ignored on dryrun, and the tempdir will get removed by the usual with-block
    cleanup.
    """
    with tarfile.open(tarball) as tar:
        tar.extractall(dest)


class Service:
    """ Cross-platform wrapper for system service operations """
    # Most of these functions work by looking up appropriate commands
    # based on sysinfo. I feel like maybe it would be better to make
    # Service an ABC with a subclass for each platform, though.
    def __init__(self, name, sysinfo=None):
        self.name = name
        self.sys = sysinfo or SysInfo()

    def _do_cmd(self, action):
        """ Get appropriate command to perform an action """
        cmds_manage = {"systemctl": "systemctl {action} {name}",
                       "service":   "service {name} {action}",
                       None:        "/etc/init.d/{name} {action}"}
        cmd = cmds_manage[self.sys.svcmgr]
        return cmd.format(name=self.name, action=action)

    def _do(self, action, suppress=None):
        """ Perform a service operation

        `suppress` is an optional list of exception types to ignore.
        Most commonly this will be CalledProcessError, when you don't
        care about the error code.
        """
        if not suppress:
            suppress = tuple()
        elif issubclass(suppress, Exception):
            suppress = (suppress,)
        try:
            return run(self._do_cmd(action))
        except suppress as ex:  # pylint: disable=catching-non-exception
            log.debug(ex)
            return None

    @property
    @lru_cache(128)
    def path(self):
        """ Get the filesystem path for the service file """
        paths = {'systemd': '/etc/systemd/system/{}.service',
                 'init': '/etc/init.d/{}'}
        name = self.name
        init = self.sys.init
        try:
            return paths[init].format(name)
        except KeyError as ex:
            msg = ("can't locate {} service path; unknown "
                   "init/service type '{}'")
            raise ValueError(msg.format(name, init)) from ex

    @property
    def installed(self):
        """ Check if the service is installed """
        return os.path.exists(self.path)

    @property
    def enabled(self):
        """ Check if the service is enabled """
        cmds = {"systemctl": "systemctl is-enabled {}",
                "chkconfig": "chkconfig {}",
                # There has to be a better way to do these two...
                "rc-update": "rc-update show | grep {}",
                "update-rc.d": "ls /etc/rc?.d/S??{}"}  # Ew
        try:
            run(cmds[self.sys.svccfg].format(self.name))
        except CalledProcessError as ex:
            log.debug(ex)
            return False
        return True

    @property
    def running(self):
        """ Check if the service is running """
        try:
            self._do('status')
        except CalledProcessError as ex:
            log.debug(ex)
            return False
        return True

    def enable(self):
        """ Enable the service """
        if self.enabled:
            log.debug("not enabling %s (already enabled)", self.name)
            return
        cmds = {"systemctl": "systemctl enable {}",
                "chkconfig": "chkconfig --add {}",
                "rc-update": "rc-update add {}",
                "update-rc.d": "update-rc.d {} defaults"}
        run(cmds[self.sys.svccfg].format(self.name))

    def disable(self):
        """ Disable the service """
        if not self.enabled:
            log.debug("not disabling %s (already disabled)", self.name)
            return
        cmds = {"systemctl": "systemctl disable {}",
                "chkconfig": "chkconfig --del {}",
                "rc-update": "rc-update del {}",
                "update-rc.d": "update-rc.d -f {} remove"}
        run(cmds[self.sys.svccfg].format(self.name))

    def start(self):
        """ Start the service """
        if self.running:
            log.debug("not starting %s (already running)", self.name)
            return
        self._do('start')

    def stop(self):
        """ Stop the service """
        if not self.running:
            log.debug("not stopping %s (already running)", self.name)
            return
        self._do('stop')

    def restart(self):
        """ Restart the service """
        self._do('restart')

    def install(self, svcdir):
        """ Install the service using the appropriate file under svcdir

        File names should follow the pattern '{inittype}.{platform}', where
        both are taken from SysInfo. A platform of 'any' matches any platform.
        If a file matching the current platform is found, it is copied to the
        appropriate location for systemd units or init scripts. Otherwise,
        OperationFailed is raised.
        """
        log.info("installing service script for %s on %s at %s",
                 self.name, self.sys.platform, self.path)
        for platform in (self.sys.platform, 'any'):
            path = '{}/{}.{}'.format(svcdir, self.sys.init, platform)
            if not os.path.exists(path):
                log.debug("no script at %s", path)
                continue
            # Older systemds don't seem to like symlinked units, so copy
            # instead of linking.
            log.debug("found script at %s", path)
            copy(path, self.path)
            if self.sys.init == 'systemd':
                run("systemctl daemon-reload")
            return
        raise OperationFailed(  # should this be FileNotFoundError?
            "failed to install {} service; no service file matches "
            "the current platform ({}:{})".
            format(self.name, self.sys.init, self.sys.platform)
            )

    def uninstall(self):
        """ Uninstall the service """
        log.info("uninstalling service script for %s on %s at %s",
                 self.name, self.sys.platform, self.path)
        self.stop()
        self.disable()
        remove(self.path)
        if self.sys.init == 'systemd':
            run("systemctl daemon-reload")


class InstallationContext:
    """ Context manager for installation operations

    Keeps track of package, install, and configuration directories, and
    provides installation operations as methods.

    package_dir: The directory where package contents live, e.g. /opt/cbu
    install_prefix: The installation path, e.g. /usr/local or ~/.local
    conf_dir: The base configuration path, e.g. /etc or ~/.config
    """
    def __init__(self, repo, package_dir, install_prefix, conf_dir):
        self.repo = repo
        self.root = normpath(expanduser(package_dir))
        self.prefix = normpath(expanduser(install_prefix))
        self.etc = normpath(expanduser(conf_dir))
        self.host = SysInfo()

    @classmethod
    def from_config(cls, conf, *args, **kwargs):
        """ Create an InstallationContext based on the config file """
        repo = Repository(conf.repo, conf.cache)
        return cls(repo, conf.unpack, conf.install, conf.etc, *args, **kwargs)

    def installed(self, name):
        """ Get installed version of a package """
        # The full path is necessary here. It makes sure we're getting the user
        # install (if running as user) or the system install (if running as
        # root), and it avoids issues with cron (which sometimes(?) doesn't
        # include /usr/local/bin in $PATH)
        cmd = "{}/{name}/bin/{name} --version".format(self.root, name=name)
        return Version.from_command(cmd, None)

    def is_installed(self, pkg):
        """ Check if a package is already installed """
        return pkg.version == self.installed(pkg.name)

    def _install_service(self, pkg, content):
        """ Install a package's system service """
        if not content.svcfiles:
            log.info("not installing services (none included in package)")
            return
        svc = Service(pkg.name)
        svcdir = '{}/{}/service'.format(self.root, pkg.name)
        try:
            svc.install(svcdir)
            svc.enable()
        except (OperationFailed, CalledProcessError) as ex:
            msg = ("Error during service setup. "
                   "{} is installed, but probably not running. ({})")
            raise OperationFailed(msg.format(pkg.name, ex)) from ex

    def _run_control_script(self, pkgname, script):
        """ Run a package control script

        If the package does not provide the specified script, nothing happens.
        If it exists but returns non-zero, OperationFailed is raised.

        The script working directory is set to the package directory, so it can
        find its own files.
        """
        path = os.path.join(self.root, pkgname, 'scripts', script)
        log.debug("looking for %s at %s", script, path)
        if not os.path.exists(path):
            log.debug("no %s script for %s, skipping", script, pkgname)
            return
        log.info("running %s script for %s", script, pkgname)
        try:
            output = run(path, cwd=os.path.join(self.root, pkgname))
        except CalledProcessError as ex:
            msg = ("{} script failed; aborting ({})")
            raise OperationFailed(msg.format(pkgname, ex)) from ex
        for line in output.decode().splitlines():
            line = line.strip()
            if line:
                log.info(line)

    def install(self, pkg):
        """ Install a package

        The tarball is unpacked to {pkg_directory}/{name}_{version}, then moved
        to {pkg_directory}/{name}.

        If a `share` subdir exists, {install_prefix}/share/{name} is symlinked
        to that directory. If there are files under `bin`, they are symlinked
        from {prefix}/bin. Files under `etc` are copied to the user or system
        conf directory *iff* they don't already exist.

        Any pre/post inst scripts under `hooks` are run at the appropriate
        times.
        """
        # Implementation note: We unpack the package, record its content,
        # and then all other operations refer to that rather than directly
        # querying the package directory. This allows dryrun to continue (not)
        # doing things even if the installed package directory is not installed
        # (because dryrun)
        #
        # The unpack itself is done regardless of dryrun; it just gets deleted
        # without being installed.
        #
        # TODO: a smarter implementation would have PackageContent get its
        # information from the tarball's content listing without actually
        # unpacking it. Though that might incur the same decompression overhead
        # as unpacking, I'm not sure.

        # Download and unpack the package
        tarball = self.repo.download(pkg)
        pkgdir = os.path.join(self.root, pkg.name)
        with TemporaryDirectory() as tmp:
            log.info("extracting: %s -> %s", tarball, tmp)
            unpack(tarball, tmp)
            content = PackageContent(tmp)
            move(tmp, pkgdir)
            # On older pythons, tempdirectory will barf on cleanup if the dir
            # it was managing no longer exists
            os.mkdir(tmp)

        # Run preinst script if required.
        if os.geteuid() == 0:
            self._run_control_script(pkg.name, "preinst")

        # Create links
        for filename in content.binfiles:
            tgt = os.path.join(pkgdir, 'bin', filename)
            src = os.path.join(self.prefix, 'bin', filename)
            link(src, tgt)
        if content.share:
            tgt = os.path.join(pkgdir, 'share')
            src = os.path.join(self.prefix, 'share', pkg.name)
            link(src, tgt)

        # Copy etc files to the config dir iff they don't already exist
        #
        # FIXME: shutil.copy preserves ownership and permissions, but that does
        # depend on them being correct in the tarball, and I don't know how it
        # will handle unexpected ownership/permissions or the user/root
        # distinction. Needs more work.
        for rpath in content.conffiles:
            src = os.path.join(pkgdir, 'etc', rpath)
            tgt = os.path.join(self.etc, rpath)
            if os.path.exists(tgt):
                log.debug("skipping existing config file: %s", tgt)
            else:
                log.debug("installing new config file: %s", tgt)
                copy(src, tgt)

        if os.geteuid() == 0:
            log.debug("we are root; running rc scripts and service install")
            self._install_service(pkg, content)
            self._run_control_script(pkg.name, "postinst")
        else:
            log.info("skipping control scripts and service install "
                     "(because we are not root)")

    def uninstall(self, name):
        """ Remove a package

        By default: removes the bin/share symlinks, then removes the package
        directory itself. Does *not* remove config files; to do that, use
        purge instead.
        """

        # TODO: make packages somewhat self-documenting, with an internal
        # metadata file stating package format version (0 old, 1 new, 2 future,
        # etc) and package version.
        if not self.installed(name):
            log.info("%s isn't installed; nothing to do", name)
            return
        if os.geteuid() == 0:
            log.info("removing %s service", name)
            Service(name).uninstall()
            self._run_control_script(name, "prerm")
        else:
            log.info("skipping control scripts and service removal "
                     "(because we are not root)")
        content = PackageContent(os.path.join(self.root, name))
        ct_links = len(content.binfiles) + int(bool(content.share))
        log.info("removing %s symlink(s) installed by %s", ct_links, name)
        for filename in content.binfiles:
            remove(os.path.join(self.prefix, 'bin', filename))
        if content.share:
            remove(os.path.join(self.prefix, 'share', name))
        if os.geteuid() == 0:
            self._run_control_script(name, "postrm")
        rmtree(content.path)


@total_ordering
class Version:
    """ Semantic versioning object

    Allows querying, parsing, and comparing versions easily. There's an
    existing package for this that's probably better, but it's not in stdlib.

    Supports a subset of the version components defined in PEP440,
    specifically those commonly generated by setuptools-scm: major, minor,
    patch, release candidate, dev build, date, and hash.
    """
    vpat = r"""
        (?:^| )
        v?
        (?P<major>\d+)\.
        (?P<minor>\d+)\.
        (?P<patch>\d+)
        (?:[-_\.]?rc[-_\.]?(?P<rc>\d+))?
        (?:[-_\.]?dev[-_\.]?(?P<dev>\d+))?
        (?:\+
            (?:g(?P<rev>[0-9a-f]{7}))?
            \.?
            (?:d(?P<date>[0-9]{8}))?
        )?
        (?:$| )
        """
    regex = re.compile(vpat, re.VERBOSE)

    def __init__(self, auto=None, major=None, minor=None, build=None,
                 rc=None, dev=None, rev=None, date=None):
        if auto and any([major, minor, build, rc, dev, rev, date]):
            raise ValueError("provide auto or version specifiers, not both")
        if isinstance(auto, str):
            match = self.regex.search(auto)
            if not match:
                raise ValueError("Invalid version specifier: " + auto)
            auto = match.groups()
            major, minor, build, rc, dev, rev, date = match.groups()
        elif isinstance(auto, Version):
            auto = (auto.major, auto.minor, auto.build,
                    auto.rc, auto.dev, auto.rev, auto.date)
        if auto:
            major, minor, build, rc, dev, rev, date = auto
        if not all((major, minor, build)):
            msg = "must provide at least major, minor, and build components"
            raise ValueError(msg)
        self.major = int(major)
        self.minor = int(minor)
        self.build = int(build)
        self.rc = rc and int(rc)
        self.dev = dev and int(dev)
        self.rev = rev
        self.date = date

    @property
    def is_release(self):
        """ Check if this is a release version """
        return not any((self.rc, self.dev, self.rev, self.date))

    def __hash__(self):
        return hash((self.major, self.minor, self.build,
                     self.rc, self.dev, self.rev, self.date))

    def __str__(self):
        release = 'v{}.{}.{}'.format(self.major, self.minor, self.build)
        rc = self.rc and 'rc{}'.format(self.rc)
        dev = self.dev and '.dev{}'.format(self.dev)
        rev = self.rev and 'g{}'.format(self.rev)
        date = self.date and 'd{}'.format(self.date)

        public = ''.join(part for part in (release, rc, dev) if part)
        local = '.'.join(part for part in (rev, date) if part)
        out = '+'.join(part for part in (public, local) if part)
        return out

    def __format__(self, spec):
        # TODO: It would be nice to support a spec option to omit or include
        # the `v` prefix, but I'm not sure how to do that in a way that will
        # still do the right thing if used to format None.
        return str(self).__format__(spec)

    def __repr__(self):
        return 'Version({})'.format(str(self))

    def _order(self):
        """ Ordering key for comparisons """
        return (self.major, self.minor, self.build,
                not self.dev,   # dev builds before non-dev
                not self.rc,    # release candidates before releases
                not self.date,  # dirty builds before clean ones
                self.rc, self.dev, self.date, self.rev)
        # Rev does not technically have a meaningful ordering, but is necessary
        # to include for equality purposes.

    def __eq__(self, other):
        return other is not None and self._order() == other._order()

    def __lt__(self, other):
        return self._order() < other._order()
    # remaining comparison functions provided by total_ordering

    @classmethod
    def from_command(cls, cmd, default=NO_DEFAULT):
        """ Run a command and parse it for a version specifier

        If `default` is passed, returns default if the command fails.

        Typically called as `Version.from_command(pkgname + ' --version')`
        """
        # FIXME: error handling here isn't great. Had a bug where something was
        # passing '{command} {version} --version', which should have generated
        # an exception but didn't.
        _run = partial(subprocess.check_output,
                       shell=True,
                       stderr=subprocess.STDOUT)
        log.debug("getting version from cmd: %s", cmd)
        catchable = (tuple() if default is NO_DEFAULT
                     else subprocess.CalledProcessError)
        try:
            return cls(_run(cmd).decode())
        except catchable as ex:
            log.debug("Couldn't parse version from %s (%s)", cmd, ex)
            return None


@total_ordering
class PackageSpec:
    """ Class representing package metadata

    Intended as an interface to PACKAGE_INDEX entries, e.g.
    `PackageSpec(*line.split())`
    """
    regex = re.compile(r'(?P<name>[\w\-]+)-'
                       r'v?(?P<version>\d+\.\d+\.\d+.*)'
                       r'\.tar\.(?:gz|bz2)')

    def __init__(self, name, version, filename, size, sha):
        self.name = name.strip()
        self.version = Version(version)
        self.filename = filename.strip()
        self.size = int(size)
        self.sha = sha

        if not self.name:
            raise ValueError("package name must be non-empty")
        if not self.filename:
            raise ValueError("package filename must be non-empty")
        if not self.size:
            raise ValueError("invalid package size: " + str(self.size))

    def __str__(self):
        return '{} {}'.format(self.name, self.version)

    def __eq__(self, other):
        return (other is not None
                and self.name == other.name
                and self.version == other.version)

    def __lt__(self, other):
        if not isinstance(other, type(self)):
            fmt = "can't order instances of '{}' and '{}'"
            raise TypeError(fmt.format(type(self), type(other)))
        return self.name < other.name or self.version < other.version

    @property
    def format(self):
        """ Get package format (i.e. file extension) """
        return splitext(self.filename)[1][1:]

    @property
    def can_unpack(self):
        """ Check whether we can unpack the package

        This may not be possible for e.g. bz2 packages on ancient systems.
        """
        return self.format in available_compression_formats()

    def validate(self, path):
        """ Validate a downloaded package file """
        log.debug("validating hash of %s", path)
        if self.sha != filehash(path):
            raise ValidationError('sha256 hash', self.sha, filehash(path))
        log.debug("hash check OK (%s)", self.sha)

    @classmethod
    def from_file(cls, path):
        """ Infer package metadata from a package file """
        filename = basename(path)
        match = cls.regex.fullmatch(filename)
        if not match:
            raise InvalidPackageFilename(filename)
        name, version = match.groups()
        return cls(name, version, filename, getsize(path), filehash(path))


class PackageContent:  # pylint: disable=too-few-public-methods
    """ Helper attributes for package directories """
    def __init__(self, path):
        self.path = path
        self.etc = self._subpath('etc')
        self.bin = self._subpath('bin')
        self.share = self._subpath('share')
        self.svc = self._subpath('service')
        self.scripts = self._subpath('scripts')
        self.binfiles = self._binfiles()
        self.conffiles = list(self._conffiles())
        self.svcfiles = list(self._svcfiles())

    def _subpath(self, sp):
        path = os.path.join(self.path, sp)
        return path if os.path.exists(path) else None

    def _binfiles(self):
        """ List binaries in the package """
        return os.listdir(self.bin) if self.bin else []

    def _svcfiles(self):
        return os.listdir(self.svc) if self.svc else []

    def _conffiles(self):
        """ List configuration file paths, relative to self.etc """
        if not self.etc:
            return
        for dirpath, _, files in os.walk(self.etc):
            for filename in files:
                yield relpath(os.path.join(dirpath, filename), self.etc)


class Repository:
    """ Interface to the CBU package repository """

    fn_index = 'PACKAGE_INDEX.txt'
    fn_signature = 'PACKAGE_INDEX.sig'

    def __init__(self, root, cache):
        parts = urlparse(root)
        if not parts.scheme:
            root = "file://" + abspath(expanduser(root))
            parts = urlparse(root)
        self.root = parts.geturl()
        self.is_local = parts.scheme == 'file'
        self.url_index = os.path.join(root, self.fn_index)
        self.url_sig = os.path.join(root, self.fn_signature)
        self.cache = expanduser(cache or Config().cache)

        try:
            os.makedirs(self.cache, 0o750)
            log.info("cache directory created at %s", self.root)
        except FileExistsError:
            log.debug("cache directory present, skipping creation")
        # FIXME: more sanity checks, e.g. verify dir perms

    def __repr__(self):
        return "Repository('{}')".format(self.root)

    def __str__(self):
        return "Repository({})".format(self.root)

    def __iter__(self):
        return iter(set(pkg.name for pkg in self.packages))

    def __getitem__(self, name):
        return (package for package in self.packages if package.name == name)

    def byname(self):
        """ Get package listing grouped by package name """
        return groupby(sorted(self.packages), key=lambda pkg: pkg.name)

    # operations requiring network requests should be done lazily and only
    # once.
    @property
    @lru_cache(128)
    def signature(self):
        """ Get index signature """
        with urlopen(self.url_sig) as response:
            return response.read().decode('utf-8')

    def index(self):
        """ Iterate over index file lines

        Empty lines are omitted, and the trailing newline is stripped.
        """
        log.debug("retrieving package index from %s", self.url_index)
        try:
            with urlopen(self.url_index) as response:
                for line in io.TextIOWrapper(response, 'utf-8'):
                    if line.strip():
                        yield line.rstrip('\n')
        except URLError as ex:
            # HTTPErrors provide the url that failed, but base URLErrors don't
            # for some stupid reason. It *should* always be url_index.
            url = getattr(ex, 'url', self.url_index)
            # Follow the reason attribute to get the root cause
            reason = ex
            while isinstance(getattr(reason, 'reason', None), Exception):
                reason = ex.reason
            msg = "failed to retrieve package index from {} ({})"
            raise BrokenRepo(msg.format(url, reason)) from ex

    def getpkg(self, name, version=None, include_prereleases=False):
        """ Get a packagespec for a given name and version

        May take either a 'name=version' string or a name and Version.

        When a package is available in multiple formats, selects the format
        with the smallest download from among those supported by the running
        system. Old systems may not support bz2/lzma.
        """
        if '=' in name:
            name, version = name.split('=')
            version = Version(version)
        # Get only the packages we know how to work with. Default to the latest
        # version out of those we can unpack, then filter the list to just that
        # version. Choose the smallest package of those that remain.
        pkgs = [pkg for pkg in self[name]
                if pkg.can_unpack
                and (pkg.version.is_release or include_prereleases)
                and (not version or pkg.version == version)]
        if not pkgs:
            raise OperationFailed(
                "can't find any packages for {} {} "
                "that we know how to open".format(name, version or '')
                )
        version = version or max(pkgs).version
        pkgs = [p for p in pkgs if p.version == version]
        log.debug("found %s pkgs for %s %s", len(pkgs), name, version)

        # FIXME: should prefer formats in cache over smaller ones
        smallest = min(pkgs, key=lambda p: p.size)
        log.debug("selected: %s (%s bytes)", smallest.filename, smallest.size)
        return smallest

    @property
    @lru_cache(128)
    def packages(self):
        """ Get repository package list """
        return [PackageSpec(*line.split()) for line in self.index()]

    def download(self, pkg, path=None, force=False):
        """ Download a package

        Unless an alternate path is specified, it will be downloaded to the
        cache directory. If the package is already present in cache, download
        will be skipped, unless force is given.

        Returns the full path to the downloaded file.
        """
        source = os.path.join(self.root, pkg.filename)
        dest = path or os.path.join(self.cache, pkg.filename)
        mode = 'wb' if force else 'xb'

        try:
            sha = filehash(dest)
        except FileNotFoundError:
            sha = None

        if force:
            log.info("unconditionally downloading %s (forced)", pkg.filename)
        elif sha == pkg.sha:
            log.info("skipping download of %s (cached)", pkg.filename)
            return dest
        elif sha is not None:
            log.debug("%s: %s (actual) != %s (expected)",
                      pkg.filename, sha, pkg.sha)
            log.info("re-downloading cached %s (hash mismatch)", pkg.filename)
            mode = 'wb'
        else:
            log.info("downloading %s -> %s", source, dest)

        with open(dest, mode) as f, urlopen(source) as response:
            f.write(response.read())
        pkg.validate(dest)
        return dest

    def clear_cache(self, force=False):
        """ Empty the cache """
        paths = [os.path.join(self.cache, filename) for filename
                 in os.listdir(self.cache)]
        if not paths:
            eprint("cache is empty; nothing to do")
            return
        if not force:
            eprint("files to be removed:")
            eprint('\n'.join(paths))
            if not Action.confirm("continue?"):
                abort("user declined confirmation prompt; aborting")
        for path in paths:
            try:
                os.remove(path)
                eprint('removed ' + path)
            except (FileNotFoundError, IsADirectoryError) as ex:
                eprint(ex)


class Config(ConfigParser):  # pylint: disable=too-many-ancestors
    """ ConfigParser that autoloads defaults """

    _global_defaults = dedent("""
        # Global settings
        [cbau]
        log_level = info
        log_file = /var/log/cbu-updater.log
        log_format = %(asctime)s\t%(levelname)s\t%(module)s:%(lineno)s\t%(message)s
        log_datefmt = %Y-%m-%dT%H:%M:%S
        repo = https://agentrepo.drivesrvr.com
        cache = /var/cache/cbu/pkgs
        unpack = /opt/cbu
        install = /usr/local
        etc = /etc
        """).strip()

    _user_defaults = dedent("""
        # Global settings
        [cbau]
        log_file = ~/.cache/cbu/log/cbu-updater.log
        cache = ~/.cache/cbu/pkgs
        unpack = ~/.local/opt/cbu
        install = ~/.local
        etc = ~/.config
        """).strip()

    system_conf = '/etc/cbu/cbau.conf'
    user_conf = '~/.config/cbu/cbau.conf'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, interpolation=None, **kwargs)
        log.debug("loading config defaults")
        self.read_string(self._global_defaults)
        if os.geteuid() != 0:
            log.debug("loading non-root user defaults")
            self.read_string(self._user_defaults)
        # Remember input file paths, so the confdump command can report what
        # was loaded.
        self.files_read = []

    # Convenience. Dot lookups do a get against the [cbau] section.
    def __getattr__(self, attr):
        return self['cbau'][attr]

    def read(self, filenames, *args, **kwargs):
        found = super().read(filenames, *args, **kwargs)
        for path in filenames:
            self.files_read.append((path, path in found))

    def dump(self, fobj, *args, head=True, **kwargs):
        """ Dump the current config with an explanatory header """
        foprint = partial(print, file=fobj)
        if head:
            foprint("# Config file search paths")
            foprint("# d/r/m = defaults/read/missing")
            foprint("#")
            foprint("# * [d] (builtin defaults)")
            for path, found in self.files_read:
                state = 'r' if found else 'm'
                foprint("# * [{}] {}".format(state, path))
            foprint("#\n# Resulting configuration dump follows:\n")
        self.write(fobj, *args, **kwargs)

    @classmethod
    def load(cls, *additional_conf_files):
        """ Read the appropriate config files and return the configuration """
        conf = cls()
        paths = chain([cls.system_conf, cls.user_conf], additional_conf_files)
        conf.read([expanduser(path) for path in paths if path])
        return conf


class SysInfo:
    """ Container for information about the running system """

    MID_FILE = '/etc/machine-id'

    def __init__(self):
        self.hostname = socket.gethostname()

    def __str__(self):
        return 'SysInfo(fingerprint={})'.format(self.fingerprint)

    # dictalike methods for ease of iteration
    def __iter__(self):
        for k in vars(self):
            yield k
        for name, obj in inspect.getmembers(type(self)):
            if isinstance(obj, property):
                yield name

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError as ex:
            raise KeyError(key) from ex

    def items(self):
        """ Get sysinfo attributes as key/value pairs

        Dictalike interface. Intended for ease of iteration
        """
        for k in self:
            yield k, self[k]

    @property
    @lru_cache(128)
    def mid(self):
        """ The machine ID, from /etc/machine-id

        Returns None on systems that predate that convention.
        """
        try:
            # pylint: disable-next=unspecified-encoding
            with open(self.MID_FILE) as f:
                return f.read().strip()
        except FileNotFoundError as ex:
            log.warning("couldn't get machine ID, using hwaddr instead (%s)",
                        ex)
            return None

    @property
    @lru_cache(128)
    def dmi_uuid(self):
        """ Get the system-uuid from dmidecode """
        sp = subprocess  # convenience alias
        try:
            cmd = 'dmidecode -s system-uuid'
            output = sp.check_output(cmd, shell=True, stderr=sp.STDOUT)
        except subprocess.CalledProcessError as ex:
            log.debug("%s", ex)
            log.info("Failed to get system uuid (dmidecode returned %s), "
                     "you may need root", ex.returncode)
            for line in ex.output.decode().splitlines():
                log.debug('dmid output: %s', line)
            return 'unknown (query failed; permission issue?)'
        return output.decode().strip()

    @property
    @lru_cache(128)
    def ip(self):
        """ Get the default IP address

        Credit to: https://stackoverflow.com/a/28950776/4638839 for the method.
        """
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
            sock.settimeout(0)
            sock.connect(('200.255.255.255', 1))
            return sock.getsockname()[0]

    @property
    def python(self):
        """ Get running python version """
        return python_version()

    @property
    def fingerprint(self):
        """ Generate system fingerprint

        The fingerprint is intended to be deterministic and unique to the
        machine. At the moment is is the sha1 hash of the system's hostname,
        ip, and machine-id, concatenated together.
        """
        components = [
            self.hostname,
            self.ip,
            self.mid,
            ]
        log.debug("fingerprint components: %s", components)
        fp = hashlib.sha1()   # pylint: disable=invalid-name
        for component in components:
            fp.update(str(component).encode())
        return fp.hexdigest()

    @property
    def init(self):
        """ Identify init type

        Mostly useful for identifying systemd systems. Returns the content
        of /proc/1/comm.
        """
        try:
            with open('/proc/1/comm', encoding='utf-8') as f:
                return f.read().strip()
        except IOError:
            return "unknown (shouldn't happen)"

    @property
    @lru_cache(128)
    def platform(self):
        """ Identify the distribution

        For now this roughly matches the old updater's platform detection, for
        compatibility's sake. There's probably a better way, though.
        """
        # Search possible release files for patterns identifying a given
        # platform. First match wins.
        #
        # TODO: might want an actual Platform class, if we want to map
        # functions based on platform. E.g. things like defining which
        # platforms get which service scripts. But that might be better done in
        # the preinst/postinst scripts.
        #
        # TODO: more methods of identification. /etc/lsb-release should be
        # relatively cross-platform for recent linuxes.

        log.debug("detecting system platform")
        Platform = namedtuple('Platform', 'platform release_file pattern')
        platforms = [
            Platform('arch',    '/etc/arch-release',    '.*'),
            Platform('gentoo',  '/etc/gentoo-release',  '.*'),
            Platform('alma',    '/etc/redhat-release',  'AlmaLinux'),
            Platform('centos',  '/etc/redhat-release',  'CentOS'),
            Platform('fedora',  '/etc/redhat-release',  'fedora'),
            Platform('rocky',   '/etc/redhat-release',  'Rocky'),
            Platform('redhat',  '/etc/redhat-release',  'Red Hat'),
            Platform('debian',  '/etc/issue',           'Debian'),
            Platform('ubuntu',  '/etc/issue',           'Ubuntu'),
        ]
        release = None  # raw string from release file(s)
        # Group by path so we only try to open each release file once
        by_rfile = attrgetter('release_file')
        platforms.sort(key=by_rfile)
        for path, group in groupby(platforms, by_rfile):
            try:
                with open(path, encoding='utf-8') as f:
                    release = next(f).strip()
            except OSError as ex:
                log.debug('%s: skipping (%s)', path, ex)
                continue
            for plid, _, pattern in group:
                match = re.match(pattern, release)
                if match:
                    log.debug("%s: '%s' matches '%s' (matched: %s)",
                              path, release, pattern, match.group().strip())
                    return plid
                log.debug("%s: '%s' does not match '%s'",
                          path, release, pattern)
        if release is None:
            release = "release files missing/unopenable"
            log.warning("failed to identify platform; couldn't open any "
                        "release files (shouldn't happen)")
        else:
            log.debug("unknown platform string: %s", release)
        log.debug("giving up on platform detection, falling back to 'unknown'")
        return "unknown ({})".format(release)

    @property
    @lru_cache(128)
    def svcmgr(self):
        """ Get the service-management command for the current platform """
        return next((cmd for cmd in ('systemctl', 'service')
                    if shutil.which(cmd)), None)

    @property
    @lru_cache(128)
    def svccfg(self):
        """ Get the service-configuration command for the current platform """
        svccmds = ('systemctl', 'chkconfig', 'rc-update', 'update-rc.d')
        return next(cmd for cmd in svccmds
                    if shutil.which(cmd))


# INITIALIZATION FUNCTIONS


def initparser():
    """ Initialize the CLI parser """
    # FIXME: pylint does not like the length of this function, and it complains
    # about unused arguments in the cmd_whatever interface when args aren't
    # needed. Possible refactor: A Command class, subclassing ArgumentParser,
    # that sets up per-command arguments and provides a run() function?

    # Controlling option order in the help gets obnoxious; anything passed in
    # with `parents` comes ahead of those explicitly specified. So we'll just
    # do *everything* with parents, and order them appropriately.

    # pylint: disable-next=invalid-name
    Parser = partial(ArgumentParser, add_help=False)

    parent = Parser()
    group = parent.add_argument_group('global options')
    addarg = group.add_argument
    addflag = partial(addarg, action='store_true')
    addarg('-C', '--conf', metavar='FILE',
           help='alternate configuration file')
    addflag('-h', '--help', action='help', help='print this help')
    addflag('-V', '--version', help='print version')
    addflag('-q', '--quiet', help='silence warnings')
    addflag('-v', '--verbose', help='verbose output')
    addflag('-D', '--debug', help='debug output')
    addflag('--pdb', help='invoke debugger on crash')

    mode = Parser()
    group = mode.add_argument_group('command options')
    addflag = partial(group.add_argument, action='store_true')
    addflag('-i', '--interactive', help='prompt before running')
    addflag('-f', '--force', help='run without prompting')
    addflag('-n', '--dryrun', help='show what would be done')

    cron = Parser()
    group = cron.add_argument_group('command options')
    addflag = partial(group.add_argument, action='store_true')
    addflag('-e', '--enabled', help='job will be enabled')

    _list = Parser()
    group = _list.add_argument_group('command options')
    addflag = partial(group.add_argument, action='store_true')
    addflag('-T', '--tsv', help='output as tsv')
    addflag('-H', '--no-header', help='omit header')
    addflag('-a', '--available', help='list all available versions')

    inst = Parser()
    group = inst.add_argument_group('command options')
    addarg = inst.add_argument
    addflag = partial(group.add_argument, action='store_true')
    addarg('package', nargs='+', help='package name(s)')
    addflag('-u', '--update', help='update package')
    addflag('-r', '--reinstall', help='reinstall package')
    addflag('-P', '--pre', help='include prereleases in latest search')

    uninst = Parser()
    addarg = uninst.add_argument
    addarg('package', nargs='+', help='package name(s)')

    download = Parser()
    addarg = download.add_argument
    addarg('package', help='package specifier')

    clean = Parser()
    group = clean.add_argument_group('command options')
    addflag = partial(group.add_argument, action='store_true')
    addflag('-f', '--force', help="don't prompt for confirmation")

    top = Parser(
        description="Cloud backup agent updater",
        conflict_handler='resolve',
        )
    sp = top.add_subparsers(title='commands', dest='command')
    addsp = partial(sp.add_parser, add_help=False)
    addsp('install', parents=[inst, mode, parent],
          help='install or update package')
    addsp('uninstall', parents=[uninst, mode, parent],
          help='uninstall package')
    addsp('download', parents=[download, mode, parent],
          help='download package w/o install')
    addsp('check', parents=[parent], help='see if an update is available')
    addsp('list', parents=[_list, parent],
          help='list available packages')
    addsp('sysinfo', parents=[parent], help='print system information')
    addsp('cronjob', parents=[cron, parent], help='generate update cronjob')
    addsp('confdump', parents=[parent], help='print configuration')
    addsp('clean', parents=[clean, parent], help='delete local cache')
    addsp('debug', parents=[parent], help='internal debug command')
    return Parser(parents=[top, parent], description=__doc__)


# COMMAND FUNCTIONS


def cmd_debug(args, conf):
    """ Scratch debugging command """
    print('args:')
    for k, v in vars(args).items():
        print('  {:<10} {}'.format(k+':', v))
    print()
    cmd_confdump(args, conf)


def cmd_list(args, conf):
    """ List available packages and their installed versions """
    repo = Repository(conf.repo, conf.cache)
    ctx = InstallationContext.from_config(conf)
    if args.available:
        for name, pkgs in repo.byname():
            current = ctx.installed(name)
            for version, pkgs in groupby(pkgs, key=lambda p: p.version):
                formats = ', '.join(sorted(set(p.format for p in pkgs)))
                line = '{} {} ({})'.format(name, version, formats)
                if version == current:
                    line += ' (installed)'
                print(line)
        sys.exit()
    fmt = '{}\t{}\t{}\t{}' if args.tsv else '{:<3}{:<22}{:<15}{}'
    if not args.no_header:
        print(fmt.format('st', 'package', 'available', 'installed'))
        print(fmt.format('--', '-------', '---------', '---------'))
    for name, versions in repo.byname():
        latest = max(versions).version
        current = ctx.installed(name)
        status = (
            'x' if not current
            else 'u' if latest > current
            else 'i' if latest == current
            else 'e' if latest < current   # e for experimental
            else '?'  # can't happen
            )
        print(fmt.format(status,
                         name,
                         latest or 'unavailable',
                         current or 'none'))


def cmd_check(args, conf):  # pylint: disable=unused-argument
    """ Check if an update is available

    Exits with status 0 if an update is available, 1 otherwise
    """
    repo = Repository(conf.repo, conf.cache)
    ctx = InstallationContext.from_config(conf)
    update_found = False
    for name, versions in repo.byname():
        latest = max(versions).version
        current = ctx.installed(name)
        if not current or latest > current:
            log.print("update available: %s %s", name, latest)
            update_found = True
    sys.exit(0 if update_found else 1)


def cmd_confdump(args, conf):  # pylint: disable=unused-argument
    """ Print the effective configuration

    Useful for initializing a new conf file with default settings.
    """
    conf.dump(sys.stdout)


def cmd_sysinfo(args, conf):  # pylint: disable=unused-argument
    """ Print system information """
    for k, v in SysInfo().items():
        print('{:<15} {}'.format(k+':', v))


def cmd_cronjob(args, conf):  # pylint: disable=unused-argument
    """ Generate an appropriate cron file for the system """
    mpd = 24 * 60  # minutes per day
    offset_minutes = int(SysInfo().fingerprint, 16) % mpd
    hours, minutes = divmod(offset_minutes, 60)

    cmd = "flock -n {lockfile} {binpath} install --update {pkgs}".format(
       lockfile="/var/lock/cbu-update-check",
       binpath="/usr/local/bin/cbu-updater",
       pkgs="cbu-updater driveclient",
       )
    fmt = """
        # Cloud Backup agent auto-update job
        #
        # The initial schedule is set quasi-randomly based on a system
        # fingerprint, to spread out load on the CBU servers. Feel free to
        # change the schedule if needed.
        #
        {state}{m} {h} * * * {user} {cmd}
        """
    print(dedent(fmt).strip().format(
        state='# ' if not args.enabled else '',
        m=minutes,
        h=hours,
        user='root',
        cmd=cmd,
        ))


def cmd_install(args, conf):
    """ Install packages """
    ctx = InstallationContext.from_config(conf)
    for pkg in args.package:
        pkg = ctx.repo.getpkg(pkg, None, args.pre)
        current = ctx.installed(pkg.name)
        target = pkg.version

        if target == current and not args.reinstall:
            log.print("{} is already installed; no action required".format(pkg))
            continue
        if current:
            log.print("Replacing %s %s with %s", pkg.name, current, target)
        else:
            log.print("Installing %s", pkg)

        try:
            ctx.install(pkg)
        except HTTPError as ex:
            fail("Couldn't download %s: %s (%s)", pkg, ex, ex.url)
        except PermissionError as ex:
            fail("%s (are you root?)", ex)
        current = ctx.installed(pkg.name)
        if current != target:
            fail("Install of %s appeared successful, but installed version "
                 "reports as %s. Something is wrong.", pkg, current)
        log.print("%s successfully installed", pkg)


def cmd_uninstall(args, conf):
    """ Uninstall packages """
    ctx = InstallationContext.from_config(conf)
    for pkg in args.package:
        installed = ctx.installed(pkg)
        if not installed:
            log.print('%s not installed; nothing to do', pkg)
            continue
        log.print("Uninstalling %s %s", pkg, installed)
        ctx.uninstall(pkg)
        log.print("%s %s successfully removed", pkg, installed)
        if ctx.installed(pkg):
            log.warning("%s appeared to uninstall successfully, but the "
                        "executable is still in PATH. This may happen if "
                        "it was installed in both user mode and root mode.")


def cmd_download(args, conf):
    """ Download a package without installing it """
    log.setLevel(min(logging.INFO, log.level))
    repo = Repository(conf.repo, conf.cache)
    pkg = repo.getpkg(args.package)
    repo.download(pkg, force=args.force)


def cmd_clean(args, conf):
    """ Clean local cache """
    Repository(conf.repo, conf.cache).clear_cache(args.force)


def main(argv=None, parser=None, cmd_dict=None):
    """ Run agent updater """
    if argv is None:
        argv = sys.argv[1:]
    # These hooks allow the ops tools to customize the command set. It's a
    # terrible method, but so far the others have been worse.
    parser = parser or initparser()
    cmd_dict = cmd_dict or globals()

    args = parser.parse_args()
    conf = Config.load(args.conf)
    CbauLogger.configure(args, conf)
    Action.setrunmode(args)
    log.debug("$PATH: %s", os.environ['PATH'])

    if args.version:
        print(getversion())
        sys.exit(0)
    if not args.command:
        parser.print_help()
        sys.exit(0)
    try:
        func = cmd_dict['cmd_' + args.command]
    except KeyError:
        log.error("command not implemented yet: %s", args.command)
        sys.exit(1)

    try:
        func(args, conf)
    except (KeyboardInterrupt, UserAbort) as ex:
        abort(str(ex))
    except CbauError as ex:
        fail(str(ex))
    except Exception as ex:  # pylint: disable=broad-except
        # Getting here indicates a bug; probably some error condition wasn't
        # checked for. Nothing to do but scream and die, and start the debugger
        # if requested.
        log.exception(ex)
        if not args.pdb:
            sys.exit(1)
        head = "CRASH -- UNHANDLED EXCEPTION"
        msg = fill("Starting debugger post-mortem. If you got here by "
                   "accident (perhaps by trying to see what --pdb does), "
                   "you can get out with 'quit'.")
        print("\n\n{}\n\n{}\n\n".format(head, msg))
        pdb.post_mortem()
        sys.exit(1)


if __name__ == '__main__':
    main()
